<?php

namespace Skoiy\Mongo\Repository;

use Illuminate\Support\Collection;
use Skoiy\Mongo\Collection as MongoCollection;
use Skoiy\Mongo\Traits\EagerLoadable;
use Skoiy\Mongo\TypeMaps\GenericDocument;
use Skoiy\Mongo\TypeMaps\In;

class EagerLoadAuthors extends EagerLoadable
{
    private $repository;
    private $items;
    private $key = 'authors';

    /**
     * @var GenericDocument
     */
    private $document;

    /**
     * @var Collection
     */
    private $documents;

    public function __construct(Repository $repository, Collection $documents)
    {
        $this->repository = $repository;
        $this->documents = $documents;
    }

    public function process(GenericDocument $document, $value = '_id')
    {
        $this->document = $document;

        $ids = $this->getAllAuthorsIds($value);
        $this->loadItems($ids);
        $this->append($this->document);
    }

    public function getAllAuthorsIds($value)
    {
        $authors = $this->documents
            ->pluck('_authors');

        if ($authors->isEmpty()) {
            return $authors;
        }

        return $authors
            ->flatten()
            ->filter(function ($author) {
                return isset($author->id);
            })
            ->map(function ($author) {
                return $author->id;
            })
            ->unique()
            ->values()
            ->toArray();
    }

    public function loadItems(array $in)
    {
        $pipeline = [
            [
                '$match' => [
                    'data.id' => [
                        '$in' => $in
                    ]
                ]
            ]
        ];

        $result = $this
            ->repository
            ->from(MongoCollection::AUTHORS)
            ->aggregate($pipeline);

        $this->items = $result->toCollection();
    }

    /**
     * @return Collection
     */
    public function items(): Collection
    {
        return $this->items;
    }

    /**
     * @return string
     */
    public function key(): string
    {
        return $this->key;
    }

    public function append(GenericDocument $postDocument)
    {
        $l10ns = $postDocument->data->l10n;

        if (!$l10ns instanceof Collection) {
            return;
        }

        foreach ($l10ns as $l10n) {
            $authorsToAppend = $this->authorsToAdd($l10n, $postDocument);
            $l10n->add($this->key(), $authorsToAppend);
        }
    }

    private function authorsToAdd(\Skoiy\Mongo\TypeMaps\L10n $l10n, $postDocument)
    {
        // If not exist the key _authors
        if (!isset($postDocument->_authors)) {
            return collect([]);
        }

        /** @var In $authorsInPostDocument */
        $authorsInPostDocument = $postDocument->_authors;

        // Filter only authors related with this l10n
        $authorsInPostDocumentArray = $authorsInPostDocument->filter(function ($author) use ($l10n) {
            return $author->onL10nId == $l10n->id;
        })->values();

        if (!$authorsInPostDocumentArray->count()) {
            return collect([]);
        }

        $ids = $authorsInPostDocumentArray->map(function ($author) {
            return $author->id;
        })->toArray();

        return $this->filterAuthorsFromItemsById($ids);
    }

    private function filterAuthorsFromItemsById(array $ids)
    {
        return $this->items()->filter(function (\Skoiy\Mongo\TypeMaps\AuthorDocument $author) use ($ids) {
            return in_array($author->data->id, $ids);
        })->values();
    }
}
