<?php

namespace Skoiy\Mongo\Repository;

use Skoiy\Skoiy;
use MongoDB\Client;
use RuntimeException;
use MongoDB\Driver\Cursor;
use Illuminate\Support\Collection;
use Illuminate\Container\Container;
use Skoiy\Mongo\Traits\EagerLoadable;
use Skoiy\Mongo\TypeMaps\PostDocument;
use Skoiy\Mongo\TypeMaps\DossierDocument;
use Skoiy\Mongo\TypeMaps\GenericDocument;
use Skoiy\Mongo\Paginator\LengthAwarePaginator;

abstract class Repository
{
    /** @var Client */
    protected $client;

    /** @var Skoiy */
    protected $skoiy;

    /** @var string */
    protected $collectionName;

    /** @var array */
    protected $pipeline;

    /** @var Cursor */
    protected $results;

    /**
     * @var array
     * @link https://www.mongodb.com/docs/v4.2/reference/method/db.collection.aggregate/
     */
    protected $options = [
        'allowDiskUse' => true,
        'comment' => ''
    ];

    /** @var array */
    protected $relations = [];

    /** @var array */
    protected $relationsMap = [];

    /** @var array */
    protected $project = [];

    public function __construct(Container $container)
    {
        $this->skoiy = $container->get(Skoiy::class);
        $this->client = $this->skoiy->getClient();
    }

    public function setSkoiy(Skoiy $skoiy)
    {
        $this->skoiy = $skoiy;
    }

    public function getClient()
    {
        return $this->client;
    }

    public function select()
    {
        return $this;
    }

    public function from($collectionName)
    {
        $this->collectionName = $collectionName;
        $this->options['typeMap'] = $this->getTypeMap($collectionName);
        return $this;
    }

    private function getTypeMap($collectionName)
    {
        return call_user_func_array([Skoiy::class, 'typeMapFor' . $collectionName], [$collectionName]);
    }

    public function where(array $pipeline)
    {
        $this->pipeline = $pipeline;
        return $this;
    }

    public function load(array $relations)
    {
        $this->relations = $relations;
        return $this;
    }

    public function project(array $projection)
    {
        $this->project = $projection;
        return $this;
    }

    public function pipeline()
    {
        if (!$this->pipeline)
            return [];

        return $this->pipeline;
    }

    public function count()
    {
        $count = ['$count' => 'total'];
        $tmp = $this->pipeline();
        $tmp[] = $count;

        $result = $this
            ->from($this->collectionName)
            ->aggregate($tmp)
            ->toCollection();

        if (!$result->isNotEmpty())
            return 0;

        return $result
            ->first()
            ->total;
    }

    /**
     * all
     *
     * @return LengthAwarePaginator
     */
    public function all(): LengthAwarePaginator
    {
        $total = $this->count();

        if (!$total) {
            return new LengthAwarePaginator([], 0, 1, 1);
        }

        $this->appendProjectionFields();

        $data = $this
            ->from($this->collectionName)
            ->aggregate($this->pipeline());

        $perPage = $total;
        $page = 1;

        $paginator = new LengthAwarePaginator($data->toCollection(), $total, $perPage, $page);
        $paginator->setCursor($this->results);
        return $paginator;
    }

    /**
     * paginate
     *
     * @param int $page
     * @param int $perPage
     * @return LengthAwarePaginator
     */
    public function paginate(int $page = 1, int $perPage = 15): LengthAwarePaginator
    {
        if ($page <= 0) {
            throw new RuntimeException("Page argument must be greater or equal to 1");
        }

        $total = $this->count();

        if (!$total) {
            return new LengthAwarePaginator([], 0, 1, 1);
        }

        $this->appendSkipAndLimit($page, $perPage);
        $this->appendProjectionFields();

        $data = $this
            ->from($this->collectionName)
            ->aggregate($this->pipeline())
            ->eagerLoad();

        $paginator = new LengthAwarePaginator($data, $total, $perPage, $page);
        $paginator->setCursor($this->results);
        return $paginator;
    }

    private function eagerLoad()
    {
        $documents = $this->toCollection();
        if (!sizeof($this->relations)) {
            return $documents;
        }

        foreach ($documents as $document) {
            $this->processEagerLoads($document, $documents);
        }

        return $documents;
    }

    public function processEagerLoads(GenericDocument $document, Collection $documents)
    {
        /** @var EagerLoadable $relation */
        foreach ($this->relations as $relation) {
            if ($document instanceof PostDocument) {
                $value = '_in';
            } else if ($document instanceof DossierDocument) {
                $value = 'on';
            }

            $instance = $this->getRelationsInstance($relation, $documents);
            $instance->process($document, $value);
        }
    }

    public function getRelationsInstance($relation, $documents)
    {
        if (!isset($this->relationsMap[$relation])) {
            $instance = new $relation($this, $documents);
            $this->relationsMap[$relation] = $instance;
        } else {
            $instance = $this->relationsMap[$relation];
        }

        return $instance;
    }

    private function appendSkipAndLimit($page, $perPage)
    {
        $this->pipeline[] = [
            '$skip' => $perPage * ($page - 1),
        ];
        $this->pipeline[] = [
            '$limit' => $perPage
        ];
    }

    private function appendProjectionFields()
    {
        if (sizeof($this->project)) {
            $this->pipeline[] = $this->project;
        }
    }

    public function withAggregateOption($key, $value)
    {
        $this->options[$key] = $value;
    }

    /**
     * aggregate
     *
     * @param $pipeline
     * @return $this
     */
    public function aggregate($pipeline)
    {
        /** @var Cursor $results */
        $this->results = $this
            ->client
            ->selectDatabase($this->skoiy->resolveDatabaseName())
            ->selectCollection($this->collectionName)
            ->aggregate($pipeline, $this->options);

        return $this;
    }

    /**
     * toCollection
     *
     * @return Collection
     */
    public function toCollection(): Collection
    {
        return Collection::make(iterator_to_array($this->results));
    }
}
