<?php

namespace Skoiy;

use Exception;
use MongoDB\Client;
use Skoiy\Mongo\TypeMaps;
use Illuminate\Container\Container;
use MongoDB\Driver\Exception\InvalidArgumentException as DriverInvalidArgumentException;
use MongoDB\Driver\Exception\RuntimeException as DriverRuntimeException;
use MongoDB\Exception\InvalidArgumentException;
use Skoiy\Mongo\Collection;
use Skoiy\Mongo\Indexes\Index;

class Skoiy
{
    /** @var Container */
    private $container;

    /** @var int */
    private $idAccount;

    /** @var callable */
    private $databaseNameResolver;

    /** @var string */
    private $databaseName;

    /** @var Client */
    private $client;

    /**
     *
     * @param $idAccount
     * @throws InvalidArgumentException for parameter/option parsing errors
     * @throws DriverInvalidArgumentException for parameter/option parsing errors in the driver
     * @throws DriverRuntimeException for other driver errors (e.g. connection errors)
     */
    public function __construct($idAccount, Client $client, $container = null)
    {
        $this->container = !$container ?? Container::getInstance();
        $this->idAccount = $idAccount;
        $this->client = $client;
    }

    /**
     * @return int
     */
    public function getIdAccount(): int
    {
        return $this->idAccount;
    }

    /**
     * @return Client
     */
    public function getClient(): Client
    {
        return $this->client;
    }

    /**
     * setDatabaseNameResolver
     *
     * @param callable $callable
     * @return void
     */
    public function setDatabaseNameResolver(callable $callable)
    {
        $this->databaseNameResolver = $callable;
    }

    /**
     * databaseName
     *
     * @return string
     */
    public function resolveDatabaseName()
    {
        return ($this->databaseNameResolver)($this->idAccount);
    }

    public static function typeMapForDossiers()
    {
        return [
            'array' => 'MongoDB\Model\BSONArray',
            //'document' => Mongo\Dossier::class,
            'root' => TypeMaps\DossierDocument::class,
            'fieldPaths' => [
                'data.isoLanguages' => 'array',
                'data' => TypeMaps\Dossier::class,
                'data.l10n' => TypeMaps\L10nCollection::class,
                'data.l10n.$' => TypeMaps\L10n::class,
                'on' => TypeMaps\In::class,
            ],
        ];
    }

    public static function typeMapForPosts()
    {
        return [
            'array' => 'MongoDB\Model\BSONArray',
            //'document' => Mongo\Dossier::class,
            'root' => TypeMaps\PostDocument::class,
            'fieldPaths' => [
                'data' => TypeMaps\Post::class,
                'data.isoLanguages' => 'array',
                'data.l10n' => TypeMaps\L10nCollection::class,
                'data.l10n.$' => TypeMaps\L10n::class,
                '_in' => TypeMaps\In::class,
                '_authors' => TypeMaps\In::class,
                '_stickers' => TypeMaps\In::class,
            ],
        ];
    }

    public static function typeMapForCollections()
    {
        return [
            'array' => 'MongoDB\Model\BSONArray',
            //'document' => Mongo\Dossier::class,
            'root' => TypeMaps\CollectionDocument::class,
            'fieldPaths' => [
                'data.isoLanguages' => 'array',
                'data' => TypeMaps\Collection::class,
                'data.l10n' => TypeMaps\L10nCollection::class,
                'data.l10n.$' => TypeMaps\L10n::class,
            ],
        ];
    }

    public static function typeMapForAuthors()
    {
        return [
            'array' => 'MongoDB\Model\BSONArray',
            //'document' => Mongo\Dossier::class,
            'root' => TypeMaps\AuthorDocument::class,
            'fieldPaths' => [
                'data.isoLanguages' => 'array',
                'data' => TypeMaps\Author::class,
                'data.l10n' => TypeMaps\L10nCollection::class,
                'data.l10n.$' => TypeMaps\L10n::class,
            ],
        ];
    }

    public static function typeMapForStickers()
    {
        return [
            'array' => 'MongoDB\Model\BSONArray',
            //'document' => Mongo\Dossier::class,
            'root' => TypeMaps\StickerDocument::class,
            'fieldPaths' => [
                'data' => TypeMaps\Sticker::class,
            ],
        ];
    }

    public static function typeMapForstickers_posts()
    {
        return [
            'array' => 'MongoDB\Model\BSONArray',
            //'document' => Mongo\Dossier::class,
            'root' => TypeMaps\StickerPostDocument::class,
            'fieldPaths' => [
                'items' => 'array',
                'items.$' => \stdClass::class,
            ],
        ];
    }

    public static function typeMapForposts_related()
    {
        return [
            'array' => 'MongoDB\Model\BSONArray',
            //'document' => Mongo\Dossier::class,
            'root' => TypeMaps\PostRelatedDocument::class,
        ];
    }

    /**
     * indexes
     *
     * @return array
     */
    public function indexes()
    {
        return [
            Collection::POSTS => [
                new Index(['data.id' => 1], ['unique' => false]),
                new Index(['data.publicId' => 1], ['unique' => false]),
                new Index(['data.l10n.id' => 1], ['unique' => false]),
                new Index(['data.l10n.tags.id' => 1], ['unique' => false]),
                new Index(['data.l10n.tags.slug' => 1], ['unique' => false]),
                new Index(['data.l10n.isoLanguage' => 1], ['unique' => false]),
                new Index(['data.l10n.title' => 'text', 'data.l10n.description' => 'text'], ['unique' => false]),
                new Index(['data.l10n.metadata.title' => 1], ['unique' => false]),
                new Index(['data.l10n.metadata.url' => 1, 'data.l10n.tags.id' => 1], ['unique' => false]),
                new Index(['data.l10n.customFields.name' => 1], ['unique' => false]),
                new Index(['data.l10n.customFields.value' => 1], ['unique' => false]),
                new Index(['_in.idCollection' => 1], ['unique' => false]),
                new Index(['_in.pidCollection' => 1], ['unique' => false]),
                new Index(['_in.idDossier' => 1], ['unique' => false]),
                new Index(['_in.pidDossier' => 1], ['unique' => false]),
                new Index(['_authors.id' => 1], ['unique' => false]),
                new Index(['_authors.pid' => 1], ['unique' => false]),
                new Index(['_stickers.token' => 1], ['unique' => false]),
            ],
            Collection::POSTS_RELATED => [
                new Index(['id' => 1], ['unique' => false])
            ],
            Collection::DOSSIERS => [
                new Index(['data.id' => 1], ['unique' => false]),
                new Index(['data.publicId' => 1], ['unique' => false]),
                new Index(['data.l10n.description' => 'text'], ['unique' => false]),
            ],
            Collection::COLLECTIONS => [
                new Index(['data.id' => 1], ['unique' => false]),
                new Index(['data.publicId' => 1], ['unique' => false]),
                new Index(['data.l10n.description' => 'text'], ['unique' => false]),
            ],
            Collection::AUTHORS => [
                new Index(['data.id' => 1], ['unique' => false]),
                new Index(['data.publicId' => 1], ['unique' => false]),
                new Index(['data.l10n.description' => 'text'], ['unique' => false]),
            ],
            Collection::STICKERS => [
                new Index(['data.id' => 1], ['unique' => false])
            ],
            Collection::STICKERS_POSTS => [
                new Index(['token' => 1], ['unique' => false]),
                new Index(['stickers_posts' => 1], ['unique' => false])
            ],
        ];
    }

    /**
     * createIndexes
     *
     * @description Only create the indexes that don't exist on the collection
     * @throws Exception
     */
    public function createIndexes($databaseName)
    {
        /**
         * @var string $collectionName
         * @var Index $index
         */
        foreach ($this->indexes() as $collectionName => $indexesToCreate) {
            try {

                return $this
                    ->getClient()
                    ->selectDatabase($databaseName)
                    ->selectCollection($collectionName)
                    ->createIndexes($this->wrapIndexesInArray($indexesToCreate));

            } catch (Exception $exception) {
                throw $exception;
            }
        }
    }

    /**
     * wrapIndexesInArray
     *
     * @param $indexesToCreate
     * @return array|array[]
     */
    private function wrapIndexesInArray($indexesToCreate)
    {
        return array_map(function (Index $index) {
            return $index->toArray();
        }, $indexesToCreate);
    }
}
